#ifndef LATTICEFRAME_HH_
#define LATTICEFRAME_HH_


#include "LatticeModel.hh"

namespace biu
{
	

		/**
		 * A lattice frame is a subset of a lattice in which all points
		 * coordinates XYZ are in the range ZERO <= XYZ < FRAMESIZE.
		 * 
		 * The lattice frame provides a one-to-one mapping of the managed
		 * points to an index and vice versa.
		 * 
		 * @author Martin Mann, Sebastian Will, Andreas Richter
		 */
	class LatticeFrame : public LatticeModel
	{
	private:
		unsigned int	frameSize;		//!< the size of the frame
		unsigned int	frameSizeP2;	//!< short access for (frameSize^2)
	public:
	
		typedef int index_type;			//!< the type of the managed indices
		
			//! construction
			//! @param _latDescriptor This lattice property handler has to
		    //! be != NULL.
		    //! @param frameSize the size of the lattice frame
		LatticeFrame(	const LatticeDescriptor* const _latDescriptor, 
						const unsigned int frameSize);
						
		LatticeFrame(const LatticeFrame& toCopy);
		virtual ~LatticeFrame();

			//! Returns the current size of the lattice frame.		
		unsigned int getFrameSize() const { return frameSize; }
		
			//! Resizes the frame.
			//!
			//! All indices, created with the old frame size will be
			//! incompatible with the new lattice frame!
		void setFrameSize(const unsigned int frameSize_);
		
			//! Returns the center of the lattice frame.
		IntPoint getCenter() const  { 
			unsigned int cen = frameSize/2 ; 
			return IntPoint(cen, cen, cen); 
		}
		
			//! Returns the highest index value available in this frame
		index_type getMaxIndex() const { 
			return getIndex(IntPoint(frameSize-1,frameSize-1,frameSize-1));
		}

			//! Tests whether or not a point is in the frame or not.
		bool isInFrame(const IntPoint& point) const;
		
			//! Converts a point to its index.
			//! The Point has to be inside the frame borders 
			//! (isInFrame() == true).
		index_type getIndex(const IntPoint& point) const;
		
			//! Converts an index to its point coordinates.
			//! Assure that the index was created with the same frame
			//! size. Otherwise the convertion will lead to a differnt 
			//! point.
		IntPoint getPoint(const index_type& index) const;
		
			//! Converts a sequence of indices to the corresponding
			//! absolute move sequence.
		MoveSequence 
		indicesToAbsMoves(const std::vector<index_type> indVec) const;
		
			//! Returns the indices of the neighbor vectors, this
			//! lattice frame is based on.
		std::vector<index_type> getIndexedNeighborhood() const;
		
		bool	operator== (const LatticeFrame &lf2) const;
		bool	operator!= (const LatticeFrame &lf2) const;
		
	}; // class LatticeFrame

		typedef std::set<biu::LatticeFrame::index_type> IndexSet;
		typedef std::vector<biu::LatticeFrame::index_type> IndexVec;

	std::ostream & operator <<(std::ostream &os, IndexSet& x);
	std::ostream & operator <<(std::ostream &os, IndexVec& x);

}	// namespace biu

#endif /*LATTICEFRAME_HH_*/
